import app from ":virtual/vitebook/app";
import {
  DATA_ASSET_BASE_PATH,
  inBrowser,
  isBoolean,
  isLoadedMarkdownPage,
  isString,
  resolveDataAssetID
} from "../shared";
import {
  getContext,
  PAGE_CTX_KEY,
  ROUTER_CTX_KEY,
  SERVER_CTX_KEY
} from "./context";
import { createMemoryHistory } from "./router/history/memory";
import { Router } from "./router/Router";
import { layouts as pageLayouts } from "./stores/layouts";
import { pages } from "./stores/pages";
import { get } from "./stores/store";
async function createRouter({
  context,
  renderers
}) {
  const target = inBrowser ? document.getElementById("app") : null;
  const router = new Router({
    target,
    context,
    baseUrl: app.baseUrl,
    history: inBrowser ? window.history : createMemoryHistory()
  });
  if (inBrowser && import.meta.env.PROD) {
    const redirects = window.__VBK_REDIRECTS_MAP__ ?? {};
    for (const from of Object.keys(redirects)) {
      const to = redirects[from];
      router.addRedirect(from, to);
    }
  }
  context.set(ROUTER_CTX_KEY, router);
  addRoutes(router, get(pages));
  await Promise.all(
    app.configs.map((config) => config({ context, router, renderers }))
  );
  handleHMR(router);
  if (inBrowser) {
    await router.go(location.href, { replace: true });
    router.listen();
  }
  return router;
}
let routes = [];
function addRoutes(router, pages2) {
  for (const page of pages2) {
    const route = {
      ...page.route,
      loader: async ({ redirect }) => {
        const pageOrRedirect = await loadPage(router, page);
        return isString(pageOrRedirect) ? redirect(pageOrRedirect) : pageOrRedirect;
      },
      prefetch: async ({ url, redirect }) => {
        const redirectTo = await loadPage(router, page, { prefetch: url });
        return isString(redirectTo) ? redirect(redirectTo) : void 0;
      }
    };
    router.addRoute(route);
    routes.push(route);
  }
}
function handleHMR(router) {
  if (import.meta.hot) {
    const updatePages = (pages2) => {
      routes.forEach((route) => router.removeRoute(route));
      routes = [];
      addRoutes(router, pages2);
    };
    let pagesUpdateTimer;
    const delayedPagesUpdate = (pages2) => {
      clearTimeout(pagesUpdateTimer);
      pagesUpdateTimer = setTimeout(() => updatePages(pages2), 100);
    };
    pages.subscribe(($pages) => delayedPagesUpdate($pages));
    pageLayouts.subscribe(() => delayedPagesUpdate(get(pages)));
  }
}
async function loadPage(router, page, { prefetch = false } = {}) {
  const prefetchURL = isBoolean(prefetch) ? void 0 : prefetch;
  let pageModule;
  let pageData;
  let layouts;
  if (import.meta.env.DEV) {
    pageModule = await page.loader();
    pageData = await loadData(router, pageModule, void 0, prefetchURL);
    if (import.meta.env.DEV && isString(pageData.__redirect__)) {
      return pageData.__redirect__;
    }
    layouts = await loadPageLayouts(router, page, prefetchURL);
  } else {
    [pageModule, layouts] = await Promise.all([
      (async () => {
        const mod = await page.loader();
        pageData = await loadData(router, mod, void 0, prefetchURL);
        return mod;
      })(),
      loadPageLayouts(router, page, prefetchURL)
    ]);
  }
  const loadedPage = {
    ...page,
    $$loaded: true,
    module: pageModule,
    layouts,
    data: pageData,
    get default() {
      return pageModule.default;
    },
    get meta() {
      return isLoadedMarkdownPage(this) ? pageModule.meta : void 0;
    }
  };
  if (!prefetch) {
    const store = getContext(router.context, PAGE_CTX_KEY);
    store.__set(loadedPage);
  }
  return loadedPage;
}
function loadPageLayouts(router, page, prefetchURL) {
  const $pageLayouts = get(pageLayouts);
  return Promise.all(
    page.layouts.map(async (index) => {
      const layout = $pageLayouts[index];
      const mod = await layout.loader();
      return {
        $$loaded: true,
        ...layout,
        module: mod,
        data: await loadData(router, mod, index, prefetchURL),
        get default() {
          return mod.default;
        }
      };
    })
  );
}
async function loadData(router, module, layoutIndex, prefetchURL) {
  if (!module.loader)
    return {};
  const pathname = prefetchURL?.pathname ?? get(router.navigation).to.pathname;
  const id = resolveDataAssetID(decodeURI(pathname), layoutIndex);
  const hashId = import.meta.env.PROD && !import.meta.env.SSR ? window.__VBK_DATA_HASH_MAP__[await hashDataId(id)] : id;
  if (!hashId)
    return {};
  if (import.meta.env.SSR) {
    const { data } = getContext(router.context, SERVER_CTX_KEY);
    return data.get(hashId) ?? {};
  }
  try {
    const json = !router.started ? getDataFromScript(hashId) : await (await fetch(`${DATA_ASSET_BASE_PATH}/${hashId}.json`)).json();
    return json;
  } catch (e) {
    console.error(e);
  }
  return {};
}
function getDataFromScript(id) {
  const dataScript = document.getElementById("__VBK_DATA__");
  try {
    const content = JSON.parse(dataScript.textContent);
    return content[id] ?? {};
  } catch (e) {
  }
  return {};
}
async function hashDataId(id) {
  const encodedText = new TextEncoder().encode(id);
  const hashBuffer = await crypto.subtle.digest("SHA-1", encodedText);
  const hashArray = Array.from(new Uint8Array(hashBuffer));
  return hashArray.map((b) => b.toString(16).padStart(2, "0")).join("").substring(0, 8);
}
export {
  createRouter,
  loadData,
  loadPage,
  loadPageLayouts
};
