import {
  calcRoutePathScore,
  compareRoutes,
  inBrowser,
  isString,
  isUndefined,
  matchRoute,
  normalizeURL,
  slash
} from "../../shared";
import { getContext, ROUTE_CTX_KEY } from "../context";
import { get, writable } from "../stores/store";
class Router {
  constructor({ target, context, baseUrl, history, routes }) {
    this._started = false;
    this._currentRoute = null;
    this._routes = [];
    this._redirects = /* @__PURE__ */ new Map();
    this._historyIndexKey = "vitebook::index";
    this._scrollKey = "vitebook:scroll";
    this._scrollPositions = {};
    this._navigation = writable(null);
    this.disabled = false;
    this.baseUrl = "/";
    this._beforeNavigate = [];
    this._afterNavigate = [];
    this.target = target;
    this.context = context;
    this._history = history;
    this.baseUrl = slash(baseUrl);
    routes?.forEach((route) => {
      this.addRoute(route);
    });
    if (!inBrowser) {
      this._url = new URL("http://ssr");
      this._historyIndex = Date.now();
      return;
    }
    document.body.setAttribute("tabindex", "-1");
    this._url = new URL(location.href);
    this._historyIndex = this._history.state?.[this._historyIndexKey];
    if (!this._historyIndex) {
      this._historyIndex = Date.now();
      this._history.replaceState(
        {
          ...this._history.state,
          [this._historyIndexKey]: this._historyIndex
        },
        "",
        location.href
      );
    }
    try {
      this._scrollPositions = JSON.parse(sessionStorage[this._scrollKey]);
    } catch {
    }
    const scroll = this._scrollPositions[this._historyIndex];
    if (scroll) {
      this._history.scrollRestoration = "manual";
      window.scrollTo(scroll.left, scroll.top);
    }
  }
  beforeNavigate(hook) {
    this._beforeNavigate.push(hook);
  }
  afterNavigate(hook) {
    this._afterNavigate.push(hook);
  }
  get url() {
    return this._url;
  }
  get navigation() {
    return {
      subscribe: this._navigation.subscribe
    };
  }
  get navigating() {
    return !!get(this._navigation);
  }
  get started() {
    return this._started;
  }
  get currentRoute() {
    return this._currentRoute;
  }
  addRedirect(from, to) {
    const fromURL = normalizeURL(isString(from) ? this.buildURL(from) : from);
    const toURL = normalizeURL(isString(to) ? this.buildURL(to) : to);
    this._redirects.set(fromURL.href, toURL.href);
  }
  getRoute(pathname) {
    const url = this.buildURL(pathname);
    return matchRoute(url, this._routes);
  }
  hasRoute(pathnameOrRoute) {
    if (!isString(pathnameOrRoute)) {
      return this._routes.some((route) => route === pathnameOrRoute);
    }
    const url = this.buildURL(pathnameOrRoute);
    return !!matchRoute(url, this._routes);
  }
  addRoute(route) {
    if (!this.hasRoute(route)) {
      this._routes.push({
        ...route,
        score: route.score ?? calcRoutePathScore(route.pathname)
      });
      this._routes = this._routes.sort(compareRoutes);
    }
  }
  removeRoute(route) {
    this._routes = this._routes.filter((r) => r !== route);
  }
  buildURL(pathnameOrURL) {
    const url = !isString(pathnameOrURL) ? new URL(pathnameOrURL) : new URL(
      pathnameOrURL,
      pathnameOrURL.startsWith("#") ? /(.*?)(#|$)/.exec(location.href)[1] : getBaseUri(this.baseUrl)
    );
    return normalizeURL(url);
  }
  owns(url) {
    if (import.meta.env.SSR)
      return true;
    return url.origin === location.origin && url.pathname.startsWith(this.baseUrl);
  }
  findRoute(pathnameOrURL) {
    const url = isString(pathnameOrURL) ? this.buildURL(pathnameOrURL) : pathnameOrURL;
    if (this.owns(url)) {
      const route = matchRoute(url, this._routes);
      if (!route)
        return null;
      const pathname = decodeURI(
        slash(url.pathname.slice(this.baseUrl.length))
      );
      const query = new URLSearchParams(url.search);
      const id = `${pathname}?${query}`;
      return { ...route, id, url };
    }
    return null;
  }
  back() {
    this._history.back();
  }
  async go(path, {
    scroll,
    replace = false,
    keepfocus = false,
    state = {}
  } = {}) {
    if (isString(path) && path.startsWith("#")) {
      const hash = path;
      this._updateHash(hash);
      this._changeHistoryState(this._url, state, replace);
      await this._scroll({ scroll, hash });
      this._updateScrollPosition();
      return;
    }
    const url = this.buildURL(path);
    if (!this.disabled) {
      return this._navigate(url, {
        scroll,
        keepfocus,
        replace,
        state
      });
    }
    await this.goLocation(url);
  }
  async goLocation(url) {
    location.href = url.href;
    return new Promise(() => {
    });
  }
  async prefetch(pathnameOrURL) {
    const url = this.buildURL(pathnameOrURL);
    let redirecting = this._redirect(url, (url2) => this.prefetch(url2));
    if (redirecting)
      return redirecting;
    const route = this.findRoute(url);
    if (!route) {
      throw new Error(
        "Attempted to prefetch a URL that does not belong to this app"
      );
    }
    const redirect = this._buildRedirect(pathnameOrURL, (redirectURL) => {
      redirecting = this.prefetch(redirectURL);
    });
    await route.prefetch?.({
      route,
      url,
      redirect
    });
    if (redirecting)
      return redirecting;
  }
  _redirect(url, handle) {
    const href = normalizeURL(url).href;
    if (!this._redirects.has(href))
      return null;
    const redirectHref = this._redirects.get(href);
    const redirectURL = new URL(redirectHref);
    if (this.owns(redirectURL)) {
      return handle(redirectURL);
    } else {
      return this.goLocation(redirectURL);
    }
  }
  _buildRedirect(from, handleRedirect) {
    const fromURL = this.buildURL(from);
    return async (pathOrURL) => {
      const redirectURL = this.buildURL(pathOrURL);
      this.addRedirect(fromURL, redirectURL);
      await handleRedirect(redirectURL);
    };
  }
  async _navigate(url, {
    scroll,
    accepted,
    blocked,
    state = {},
    keepfocus = false,
    replace = false
  }) {
    let cancelled = false;
    const cancel = () => {
      this._navigation.set(null);
      if (!cancelled)
        blocked?.();
      cancelled = true;
    };
    const redirectNavigation = (url2) => this._navigate(url2, {
      keepfocus,
      replace,
      state,
      accepted,
      blocked: cancel
    });
    let redirecting = this._redirect(url, redirectNavigation);
    if (redirecting)
      return redirecting;
    if (!this.owns(url)) {
      cancel();
      return;
    }
    const route = this.findRoute(url);
    if (!route) {
      cancel();
      throw new Error(
        "Attempted to navigate to a URL that does not belong to this app"
      );
    }
    const redirect = this._buildRedirect(url, (redirectURL) => {
      redirecting = redirectNavigation(redirectURL);
    });
    for (const hook of this._beforeNavigate) {
      hook({
        from: this._currentRoute,
        to: route,
        match: route.match,
        cancel,
        redirect
      });
    }
    if (cancelled)
      return;
    if (redirecting)
      return redirecting;
    this._updateScrollPosition();
    accepted?.();
    this._navigation.set({ from: this._url, to: url });
    const page = await route.loader({ route, redirect });
    if (redirecting)
      return redirecting;
    if (!page) {
      cancel();
      return;
    }
    const fromRoute = this._currentRoute;
    const toRoute = { ...route, page };
    this._currentRoute = toRoute;
    getContext(this.context, ROUTE_CTX_KEY).__set(toRoute);
    if (inBrowser) {
      await new Promise((res) => window.requestAnimationFrame(res));
    }
    this._changeHistoryState(url, state, replace);
    if (inBrowser) {
      if (!keepfocus) {
        const root = document.body;
        const tabindex = root.getAttribute("tabindex");
        getSelection()?.removeAllRanges();
        root.tabIndex = -1;
        root.focus({ preventScroll: true });
        if (tabindex !== null) {
          root.setAttribute("tabindex", tabindex);
        } else {
          root.removeAttribute("tabindex");
        }
      }
      await new Promise((res) => window.requestAnimationFrame(res));
      await this._scroll({
        from: fromRoute,
        to: toRoute,
        scroll,
        hash: url.hash
      });
    }
    await Promise.all(
      this._afterNavigate.map(
        (hook) => hook({
          from: fromRoute,
          to: toRoute,
          match: route.match
        })
      )
    );
    this._url = url;
    this._navigation.set(null);
  }
  _changeHistoryState(url, state, replace) {
    const change = replace ? 0 : 1;
    state[this._historyIndexKey] = this._historyIndex += change;
    this._history[replace ? "replaceState" : "pushState"](state, "", url);
  }
  _updateHash(hash) {
    this._url.hash = hash;
    if (this._currentRoute) {
      getContext(this.context, ROUTE_CTX_KEY).__set({
        ...this.currentRoute,
        url: this._url
      });
    }
  }
  _updateScrollPosition() {
    if (!inBrowser)
      return;
    this._scrollPositions[this._historyIndex] = scrollState();
  }
  async _scroll({
    scroll,
    hash,
    from,
    to
  }) {
    if (!inBrowser)
      return;
    let cancelled = false;
    const cancel = () => {
      cancelled = true;
    };
    let scrollTarget = null;
    if (scroll) {
      scrollTarget = await scroll({ cancel });
    } else if (!isUndefined(from) && to && this.scrollBehavior) {
      scrollTarget = await this.scrollBehavior({
        from,
        to,
        cancel,
        savedPosition: this._scrollPositions[this._historyIndex]
      });
    }
    if (scrollTarget === false) {
      cancel();
      return;
    }
    if (cancelled)
      return;
    const deepLinked = hash ? document.getElementById(decodeURI(hash).slice(1)) : null;
    const scrollTo = (options) => {
      if ("scrollBehavior" in document.documentElement.style) {
        window.scrollTo(options);
      } else {
        window.scrollTo(options.left ?? 0, options.top ?? 0);
      }
    };
    const base = this.scrollBase?.();
    const baseTop = base?.top ?? 0;
    const baseLeft = base?.left ?? 0;
    const baseBehavior = base?.behavior ?? "auto";
    if (scrollTarget || deepLinked) {
      const el = isString(scrollTarget?.el) ? document.querySelector(scrollTarget.el) : scrollTarget?.el ?? deepLinked;
      const docRect = document.documentElement.getBoundingClientRect();
      const elRect = el?.getBoundingClientRect() ?? { top: 0, left: 0 };
      const offsetTop = baseTop + (scrollTarget?.top ?? 0);
      const offsetLeft = baseLeft + (scrollTarget?.left ?? 0);
      const behavior = scrollTarget?.behavior ?? baseBehavior;
      scrollTo({
        left: elRect.left - docRect.left - offsetLeft,
        top: elRect.top - docRect.top - offsetTop,
        behavior
      });
    } else {
      scrollTo({ top: baseTop, left: baseLeft });
    }
  }
  listen() {
    this._history.scrollRestoration = "manual";
    addEventListener("beforeunload", () => {
      this._history.scrollRestoration = "auto";
    });
    addEventListener("load", () => {
      this._history.scrollRestoration = "manual";
    });
    const prefetched = /* @__PURE__ */ new Set();
    const triggerPrefetch = (event) => {
      const a = findAnchor(event.target);
      if (a && a.href && a.hasAttribute("data-prefetch") && !prefetched.has(a.href)) {
        this.prefetch(getHref(a));
        prefetched.add(a.href);
      }
    };
    let mouseMoveTimeout;
    const handleMouseMove = (event) => {
      clearTimeout(mouseMoveTimeout);
      mouseMoveTimeout = setTimeout(() => {
        triggerPrefetch(event);
      }, 30);
    };
    addEventListener("touchstart", triggerPrefetch);
    addEventListener("mousemove", handleMouseMove);
    let hashNavigation = false;
    addEventListener("click", (event) => {
      if (this.disabled)
        return;
      if (event.button || event.which !== 1)
        return;
      if (event.metaKey || event.ctrlKey || event.shiftKey || event.altKey)
        return;
      if (event.defaultPrevented)
        return;
      const a = findAnchor(event.target);
      if (!a || !a.href)
        return;
      const url = getHref(a);
      const isSvgAEl = a instanceof SVGAElement;
      if (!isSvgAEl && url.origin === "null")
        return;
      const rel = (a.getAttribute("rel") || "").split(/\s+/);
      if (a.hasAttribute("download") || rel.includes("external"))
        return;
      if (isSvgAEl ? a.target.baseVal : a.target)
        return;
      const [base, hash] = url.href.split("#");
      if (hash !== void 0 && base === location.href.split("#")[0]) {
        hashNavigation = true;
        this._updateHash(hash);
        this._updateScrollPosition();
        return;
      }
      this._navigate(url, {
        scroll: a.hasAttribute("data-noscroll") ? () => scrollState() : null,
        replace: url.href === location.href,
        keepfocus: false,
        accepted: () => event.preventDefault(),
        blocked: () => event.preventDefault()
      });
    });
    addEventListener("visibilitychange", () => {
      if (document.visibilityState === "hidden") {
        this._updateScrollPosition();
        try {
          sessionStorage[this._scrollKey] = JSON.stringify(
            this._scrollPositions
          );
        } catch {
        }
      }
    });
    addEventListener("popstate", (event) => {
      if (!event.state || this.disabled)
        return;
      if (event.state[this._historyIndexKey] === this._historyIndex)
        return;
      this._navigate(new URL(location.href), {
        scroll: () => this._scrollPositions[event.state[this._scrollKey]],
        keepfocus: false,
        accepted: () => {
          this._historyIndex = event.state[this._historyIndexKey];
        },
        blocked: () => {
          const delta = this._historyIndex - event.state[this._historyIndexKey];
          this._history.go(delta);
        }
      });
    });
    addEventListener("hashchange", () => {
      if (hashNavigation) {
        hashNavigation = false;
        this._history.replaceState(
          {
            ...this._history.state,
            [this._historyIndexKey]: ++this._historyIndex
          },
          "",
          location.href
        );
      }
    });
    this._started = true;
  }
}
function scrollState() {
  return {
    top: scrollY,
    left: scrollX
  };
}
function getHref(node) {
  return node instanceof SVGAElement ? new URL(node.href.baseVal, document.baseURI) : new URL(node.href);
}
function getBaseUri(baseUrl = "/") {
  return import.meta.env.SSR ? `http://ssr${baseUrl}` : `${location.protocol}//${location.host}${baseUrl === "/" ? "" : baseUrl}`;
}
function findAnchor(node) {
  while (node && node.nodeName.toUpperCase() !== "A")
    node = node.parentNode;
  return node;
}
export {
  Router
};
