// src/node/app/build/adapter/static.ts
function createStaticBuildAdapter(options = {}) {
  return (app, bundles, build, $) => {
    $.logger.info($.color.bold(`vitebook@${app.version}`));
    const startTime = Date.now();
    const renderingSpinner = $.createSpinner();
    return {
      name: "static",
      startRenderingPages() {
        renderingSpinner.start(
          $.color.bold(
            `Rendering ${$.color.underline(app.nodes.pages.size)} pages...`
          )
        );
      },
      finishRenderingPages() {
        renderingSpinner.stopAndPersist({
          symbol: $.icons.success,
          text: $.color.bold(
            `Rendered ${$.color.underline(app.nodes.pages.size)} pages`
          )
        });
      },
      async write() {
        let redirectsScriptTag = "";
        const redirectFiles = $.createFilesArray();
        if (!options.skipRedirects) {
          const redirectsTable = {};
          for (const redirect of build.redirects.values()) {
            redirectFiles.push({
              filename: redirect.filename,
              content: redirect.html
            });
            redirectsTable[redirect.from] = redirect.to;
          }
          const serializedRedirectsTable = JSON.stringify(redirectsTable);
          redirectsScriptTag = `<script id="__VBK_REDIRECTS_MAP__">window.__VBK_REDIRECTS_MAP__ = ${serializedRedirectsTable};<\/script>`;
        }
        const dataTable = {};
        const dataFiles = $.createFilesArray();
        for (const data of build.data.values()) {
          dataFiles.push({
            filename: data.filename,
            content: data.serializedData
          });
          dataTable[data.idHash] = data.contentHash;
        }
        const serializedDataTable = JSON.stringify(dataTable);
        const dataHashScriptTag = `<script id="__VBK_DATA_HASH_MAP__">window.__VBK_DATA_HASH_MAP__ = ${serializedDataTable};<\/script>`;
        const htmlFiles = $.createFilesArray();
        const buildingSpinner = $.createSpinner();
        const htmlPagesCount = $.color.underline(build.renders.size);
        buildingSpinner.start(
          $.color.bold(`Building ${htmlPagesCount} HTML pages...`)
        );
        const template = $.getHTMLTemplate();
        const entrySrc = bundles.client.entryChunk.fileName;
        const entryScriptTag = `<script type="module" src="/${entrySrc}" defer><\/script>`;
        const stylesheetTag = bundles.client.appCSSAsset ? $.createLinkTag("stylesheet", bundles.client.appCSSAsset.fileName) : "";
        for (const render of build.renders.values()) {
          const { assets, imports, dynamicImports } = $.resolvePageChunks(
            render.page,
            render.ssr.context.modules
          );
          const preloadLinkTags = [...assets, ...imports].map(
            (fileName) => $.createPreloadTag(fileName)
          );
          const prefetchLinkTags = dynamicImports.map(
            (fileName) => $.createLinkTag("prefetch", fileName)
          );
          const headTags = [
            render.ssr.head ?? "",
            stylesheetTag,
            ...preloadLinkTags,
            ...prefetchLinkTags
          ].filter((t) => t.length > 0).join("\n    ");
          const bodyTags = [
            redirectsScriptTag,
            dataHashScriptTag,
            $.createDataScriptTag(render.dataAssetIds),
            entryScriptTag
          ].join("");
          const pageHtml = template.replace(`<!--@vitebook/head-->`, headTags).replace(`<!--@vitebook/app-->`, render.ssr.html).replace("<!--@vitebook/body-->", bodyTags);
          htmlFiles.push({
            filename: render.filename,
            content: pageHtml
          });
        }
        buildingSpinner.stopAndPersist({
          text: $.color.bold(`Built ${htmlPagesCount} HTML pages`),
          symbol: $.icons.success
        });
        if (app.config.sitemap.length > 0) {
          const sitemapsSpinner = $.createSpinner();
          const sitemapCount = $.color.underline(app.config.sitemap.length);
          sitemapsSpinner.start(
            $.color.bold(`Building ${sitemapCount} sitemaps...`)
          );
          const sitemaps = await $.buildSitemaps();
          for (const [filename, content] of sitemaps) {
            htmlFiles.push({ filename, content });
          }
          sitemapsSpinner.stopAndPersist({
            text: $.color.bold(`Built ${sitemapCount} sitemaps`),
            symbol: $.icons.success
          });
        }
        await $.writeFiles(
          htmlFiles,
          (filename) => app.dirs.client.resolve(filename),
          (count) => `Writing ${count} HTML files`,
          (count) => `Committed ${count} HTML files`
        );
        if (redirectFiles.length) {
          await $.writeFiles(
            dataFiles,
            (filename) => app.dirs.client.resolve(filename),
            (count) => `Writing ${count} HTML redirect files`,
            (count) => `Committed ${count} HTML redirect files`
          );
        }
        if (dataFiles.length) {
          await $.writeFiles(
            dataFiles,
            (filename) => app.dirs.client.resolve(filename),
            (count) => `Writing ${count} data files`,
            (count) => `Committed ${count} data files`
          );
        }
      },
      async close() {
        $.logBadLinks();
        $.logRoutes();
        const icons = {
          10: "\u{1F92F}",
          20: "\u{1F3CE}\uFE0F",
          30: "\u{1F3C3}",
          40: "\u{1F40C}",
          Infinity: "\u26B0\uFE0F"
        };
        const endTime = ((Date.now() - startTime) / 1e3).toFixed(2);
        const formattedEndTime = $.color.underline(endTime);
        const icon = icons[Object.keys(icons).find((t) => endTime <= t)];
        $.logger.success(
          $.color.bold(`Build complete in ${formattedEndTime} ${icon}`)
        );
        const pkgManager = await $.guessPackageManager();
        const previewCommand = await $.findPreviewScriptName();
        console.log(
          $.color.bold(
            `\u26A1 ${previewCommand ? `Run \`${pkgManager === "npm" ? "npm run" : pkgManager} ${previewCommand}\` to serve production build` : "Ready for preview"}
`
          )
        );
      }
    };
  };
}

// src/node/app/http/errors.ts
function httpError(statusCode, message) {
  throw new HTTPError(statusCode, message);
}
var HTTPError = class extends Error {
  constructor(statusCode, message) {
    super(message);
    this.statusCode = statusCode;
  }
};
var HTTPJSONError = class extends Error {
  constructor(statusCode, data) {
    super();
    this.statusCode = statusCode;
    this.data = data;
  }
};
function isHTTPError(error) {
  return error instanceof HTTPError;
}
function isHTTPJSONError(error) {
  return error instanceof HTTPJSONError;
}

// src/node/app/http/http-bridge.ts
import * as set_cookie_parser from "set-cookie-parser";
async function getRequest(base, req) {
  let headers = req.headers;
  if (req.httpVersionMajor === 2) {
    headers = Object.assign({}, headers);
    delete headers[":method"];
    delete headers[":path"];
    delete headers[":authority"];
    delete headers[":scheme"];
  }
  return new Request(base + req.url, {
    method: req.method,
    headers,
    body: getRawBody(req)
  });
}
function getRawBody(req) {
  const h = req.headers;
  if (!h["content-type"]) {
    return null;
  }
  const length = Number(h["content-length"]);
  if (isNaN(length) && h["transfer-encoding"] == null) {
    return null;
  }
  if (req.destroyed) {
    const readable = new ReadableStream();
    readable.cancel();
    return readable;
  }
  let size = 0;
  let cancelled = false;
  return new ReadableStream({
    start(controller) {
      req.on("error", (error) => {
        controller.error(error);
      });
      req.on("end", () => {
        if (cancelled)
          return;
        controller.close();
      });
      req.on("data", (chunk) => {
        if (cancelled)
          return;
        size += chunk.length;
        if (size > length) {
          controller.error(new Error("content-length exceeded"));
          return;
        }
        controller.enqueue(chunk);
        if (controller.desiredSize === null || controller.desiredSize <= 0) {
          req.pause();
        }
      });
    },
    pull() {
      req.resume();
    },
    cancel(reason) {
      cancelled = true;
      req.destroy(reason);
    }
  });
}
async function setResponse(res, response) {
  const headers = Object.fromEntries(response.headers);
  if (response.headers.has("set-cookie")) {
    const header = response.headers.get("set-cookie");
    const split = set_cookie_parser.splitCookiesString(header);
    headers["set-cookie"] = split;
  }
  res.writeHead(response.status, headers);
  if (!response.body) {
    res.end();
    return;
  }
  const reader = response.body.getReader();
  if (res.destroyed) {
    reader.cancel();
    return;
  }
  const cancel = (error) => {
    res.off("close", cancel);
    res.off("error", cancel);
    reader.cancel(error).catch(() => {
    });
    if (error)
      res.destroy(error);
  };
  res.on("close", cancel);
  res.on("error", cancel);
  next();
  async function next() {
    try {
      for (; ; ) {
        const { done, value } = await reader.read();
        if (done)
          break;
        if (!res.write(value)) {
          res.once("drain", next);
          return;
        }
      }
      res.end();
    } catch (error) {
      cancel(error instanceof Error ? error : new Error(String(error)));
    }
  }
}

// src/node/app/http/request.ts
async function handleRequest(request, urlPattern, getClientAddress, loader, methods) {
  var _a;
  const methodOverride = request.method === "POST" ? (await request.formData()).get("_method") : null;
  const method = typeof methodOverride === "string" ? methodOverride : request.method;
  if (methods && !methods.includes(method))
    throw httpError(404, "Not found");
  const url = new URL(request.url);
  if (!urlPattern.test(url))
    throw httpError(404, "Not found");
  const mod = await loader();
  let handler = mod[method];
  if (!handler && method === "HEAD")
    handler = mod.GET;
  if (!handler)
    handler = mod.ANY;
  if (!handler) {
    return new Response(`${method} method not allowed`, {
      status: 405,
      headers: {
        allow: getAllowedMethods(mod).join(", ")
      }
    });
  }
  const headers = {};
  const cookies = [];
  const params = ((_a = urlPattern.exec(url)) == null ? void 0 : _a.pathname.groups) ?? {};
  const response = await handler(
    createRequestEvent(
      request,
      url,
      params,
      getClientAddress,
      headers,
      cookies
    )
  );
  if (!(response instanceof Response)) {
    throw new Error(
      `Invalid return value from route handler at ${url.pathname}. Should return a \`Response\`.`
    );
  }
  for (const name in headers) {
    const value = headers[name];
    if (value !== null)
      response.headers.set(name, value + "");
  }
  for (const cookie of cookies) {
    response.headers.append("set-cookie", cookie);
  }
  return response;
}
function createRequestEvent(request, url, params, getClientAddress, headers = {}, cookies = []) {
  return {
    get url() {
      return url;
    },
    get clientAddress() {
      const address = getClientAddress();
      if (typeof address !== "string") {
        throw new Error("Could not determine `clientAddress`");
      }
      return address;
    },
    get params() {
      return params;
    },
    get request() {
      return request;
    },
    setHeaders(newHeaders) {
      for (const key in newHeaders) {
        const name = key.toLowerCase();
        const value = newHeaders[key];
        if (name === "set-cookie") {
          const newCookies = Array.isArray(value) ? value : value ? [value + ""] : [];
          for (const cookie of newCookies) {
            if (cookies.includes(cookie)) {
              throw new Error(
                `"${key}" header already has cookie with same value`
              );
            }
            cookies.push(cookie);
          }
        } else if (name in headers) {
          throw new Error(`"${key}" header is already set`);
        } else if (typeof value !== "string") {
          throw new Error(
            `"${key}" value must be a string (found ${typeof value}).`
          );
        } else {
          headers[name] = value;
        }
      }
    }
  };
}
function getAllowedMethods(mod) {
  const allowed = [];
  for (const method of HTTP_METHODS) {
    if (method in mod)
      allowed.push(method);
  }
  if (mod.GET || mod.HEAD)
    allowed.push("HEAD");
  return allowed;
}

// src/node/app/http/response.ts
var json = (data, init = {}) => {
  const responseInit = typeof init === "number" ? { status: init } : init;
  const headers = new Headers(responseInit.headers);
  if (!headers.has("Content-Type")) {
    headers.set("Content-Type", "application/json; charset=utf-8");
  }
  return new Response(JSON.stringify(data), {
    ...responseInit,
    headers
  });
};

// src/node/app/http/http.ts
function createHTTPRequestHandler(urlPattern, loader, options) {
  let _pattern;
  const _getBase = (options == null ? void 0 : options.getBase) ?? ((req) => `https://${req.headers.host}`);
  const _getClientAddress = (options == null ? void 0 : options.getClientAddress) ?? ((req) => req.socket);
  const _handleUnknownError = (options == null ? void 0 : options.handleUnknownError) ?? ((_, res) => {
    res.statusCode = 500;
    res.end("Internal server error");
  });
  return async (req, res) => {
    var _a;
    await ((_a = options == null ? void 0 : options.installPolyfills) == null ? void 0 : _a.call(options));
    let request;
    try {
      request = await getRequest(_getBase(req), req);
    } catch (error) {
      res.statusCode = error.status || 400;
      res.end(error.reason || "Invalid request body");
      return;
    }
    try {
      const response = await handleRequest(
        request,
        _pattern ??= urlPattern(),
        () => _getClientAddress(req),
        loader,
        options == null ? void 0 : options.methods
      );
      setResponse(res, response);
    } catch (error) {
      if (isHTTPError(error)) {
        setResponse(res, handleHTTPError(error));
      } else {
        _handleUnknownError(req, res, error);
      }
    }
  };
}
function handleHTTPError(error) {
  if (isHTTPJSONError(error)) {
    return json(error.data, error.statusCode);
  } else if (isHTTPError(error)) {
    return new Response(error.message, { status: error.statusCode });
  } else {
    return new Response("Internal server error", { status: 500 });
  }
}
var HTTP_METHODS = /* @__PURE__ */ new Set([
  "GET",
  "POST",
  "PUT",
  "PATCH",
  "DELETE"
]);

export {
  createStaticBuildAdapter,
  handleRequest,
  createHTTPRequestHandler,
  HTTP_METHODS
};
