import { noslash, slash } from "./url";
const PATH_SCORE = {
  Segment: 6,
  Static: 5,
  Group: 3,
  Dynamic: 2,
  BonusRegExp: 1,
  BonusExt: 3,
  PenaltyOptional: -4,
  PenaltyWildcard: -7,
  PenaltyRepeatable: -8,
  PenaltyRegexWildcard: -10
};
const bonusRegexRE = /\(.*?\)/;
const bonusExtRE = /\.html$/;
const penaltyOptionalRE = /(^\/:(\w|\.)+\?)|({:?.*?}\?)|(\(.*?\)\?)/;
const penaltyWildcardRE = /^\/(\*|(:?\w*?\*))/;
const penaltyRepeatableRE = /^\/(:(\w|\.)+(\+|\*))|({:?.*?}(\+|\*))/;
const penaltyRegexWildcardRE = /(\(\.\*\??\)|(\.\+\??))/;
const scoreCache = /* @__PURE__ */ new Map();
function calcRoutePathScore(pathname) {
  if (scoreCache.has(pathname))
    return scoreCache.get(pathname);
  const segments = splitRoutePath(pathname);
  let score = 0;
  for (const segment of segments) {
    const isDynamic = isRoutePathDynamic(segment);
    const isNamedGroup = segment.startsWith("/:") || segment.startsWith("/{:");
    let isSegment = true;
    if (bonusExtRE.test(segment)) {
      score += PATH_SCORE.BonusExt;
    }
    if (isNamedGroup) {
      score += PATH_SCORE.Group;
    } else if (isDynamic) {
      score += PATH_SCORE.Dynamic;
    }
    if (isNamedGroup || isDynamic) {
      if (bonusRegexRE.test(segment)) {
        score += PATH_SCORE.BonusRegExp;
      }
      if (penaltyOptionalRE.test(segment)) {
        score += PATH_SCORE.PenaltyOptional;
        isSegment = false;
      }
      if (penaltyWildcardRE.test(segment)) {
        score += PATH_SCORE.PenaltyWildcard;
        isSegment = false;
      }
      if (penaltyRepeatableRE.test(segment)) {
        score += PATH_SCORE.PenaltyRepeatable;
        isSegment = false;
      }
      if (penaltyRegexWildcardRE.test(segment)) {
        score += PATH_SCORE.PenaltyRegexWildcard;
        isSegment = false;
      }
    } else {
      score += PATH_SCORE.Static;
    }
    if (isSegment)
      score += PATH_SCORE.Segment;
  }
  scoreCache.set(pathname, score);
  return score;
}
function compareRoutes(routeA, routeB) {
  if (routeA.score !== routeB.score) {
    return routeB.score - routeA.score;
  }
  const segmentsA = splitRoutePath(routeA.pathname).length;
  const segmentsB = splitRoutePath(routeB.pathname).length;
  return segmentsA != segmentsB ? segmentsB - segmentsA : routeB.pathname.length - routeA.pathname.length;
}
const trailingHtmlExtGroupRE = /{\.html}\?$/;
const trailingSlashGroupRE = /{\/}\?{index}\?{\.html}\?$/;
function cleanRoutePath(pathname) {
  return pathname.replace(trailingSlashGroupRE, "/index.html").replace(trailingHtmlExtGroupRE, ".html");
}
const routeSplitRE = /\/(?![^{(]*}|\))/g;
function splitRoutePath(pathname) {
  return noslash(cleanRoutePath(pathname)).split(routeSplitRE).map(slash);
}
const dynamicPathRE = /(\*|\(.*\)|{.*}|\/:)/;
function isRoutePathDynamic(pathname) {
  return pathname.startsWith("/:") || dynamicPathRE.test(cleanRoutePath(pathname));
}
function matchRoute(url, routes) {
  const result = matchRouteInfo(url, routes);
  return result ? { ...result.route, match: result.match } : void 0;
}
function matchRouteInfo(url, routes) {
  const normalized = "route" in routes[0] ? routes.map((r) => r.route) : routes;
  const index = normalized.findIndex(
    (route2) => route2.pattern.test({ pathname: getRouteMatchingPathname(url) })
  );
  const route = normalized[index];
  const match = execRouteMatch(url, route);
  return route && match ? { index, route, match } : void 0;
}
function execRouteMatch(url, route) {
  return route?.pattern.exec({ pathname: getRouteMatchingPathname(url) })?.pathname;
}
const htmlExtRE = /\.html$/;
function getRouteMatchingPathname(url) {
  return url.pathname.replace(htmlExtRE, "");
}
function normalizeURL(url) {
  url.pathname = url.pathname.replace(/\/index.html$/, "/");
  return url;
}
export {
  calcRoutePathScore,
  cleanRoutePath,
  compareRoutes,
  execRouteMatch,
  getRouteMatchingPathname,
  isRoutePathDynamic,
  matchRoute,
  matchRouteInfo,
  normalizeURL,
  splitRoutePath
};
